from django_glue.conf import settings
from django_glue import __version__


def glue(request):
    return {
        'DJANGO_GLUE_VERSION': __version__,
        settings.DJANGO_GLUE_CONTEXT_NAME: request.session.get(settings.DJANGO_GLUE_SESSION_NAME, {}).get('context'),
        settings.DJANGO_GLUE_KEEP_LIVE_CONTEXT_NAME: request.session.get(settings.DJANGO_GLUE_KEEP_LIVE_SESSION_NAME, {}),
        'DJANGO_GLUE_KEEP_LIVE_INTERVAL_TIME_MILLISECONDS': settings.DJANGO_GLUE_KEEP_LIVE_INTERVAL_TIME_SECONDS * 1000,
    }



import random, datetime
import uuid

from django.utils.timezone import now

from tests.models import TestModel, BigTestModel

FIRST_NAME_TUPLE = (
    'Fred',
    'Jane',
    'Janet',
    'Austin',
    'Johnny',
    'Nathan',
    'Robert',
    'Ted',
    'Wesley',
)

LAST_NAME_TUPLE = (
    'Doe',
    'Johnson',
    'Hansen',
    'Mancal',
    'Smith',
    'Waldern',
    'Howery',
    'Wilson',
    'Sauer',
)

DESCRIPTION_WORD_TUPLE = (
    'a',
    'are',
    'artsy',
    'awesome',
    'balloon',
    'boring',
    'can',
    'cat',
    'cool',
    'dripping',
    'dog',
    'eggs',
    'extra',
    'fast',
    'finished',
    'fly',
    'goal',
    'graduated',
    'handy',
    'has',
    'has',
    'have',
    'how',
    'I',
    'it',
    'junk',
    'killer',
    'loves cars',
    'mother',
    'missle',
    'need',
    'opposite',
    'panther',
    'quiz',
    'rugged',
    'she',
    'started',
    'stupid',
    'taco',
    'that',
    'then',
    'there',
    'to',
    'ugly',
    'vixen',
    'want',
    'when',
    'where',
    'when',
    'wood carving',
    'xavier',
    'you',
    'zebra',
)


def custom_test_function(request):
    return 'Hello World'


def generate_randomized_test_model(limit=5):
    new_description = ''
    for x in range(20):
        new_description += f'{DESCRIPTION_WORD_TUPLE[random.randint(0, (len(DESCRIPTION_WORD_TUPLE) - 1))]} '

    test_model_object = TestModel.objects.create(
        first_name=FIRST_NAME_TUPLE[random.randint(0, (len(FIRST_NAME_TUPLE) - 1))],
        last_name=LAST_NAME_TUPLE[random.randint(0, (len(LAST_NAME_TUPLE) - 1))],
        description=new_description,
        favorite_number=random.randint(0, 999),
        weight_lbs=round(random.uniform(80.001, 400.123), 3),
    )

    exclude_test_model = TestModel.objects.all().order_by('-id')[:limit]
    TestModel.objects.exclude(pk__in=exclude_test_model).delete()

    return test_model_object


def generate_big_test_model(limit=5):
    big_test_model = BigTestModel.objects.create(
        big_integer_field=-8294853213,
        binary_field=bytes('Some Bytes', 'utf-8'),
        boolean_field=False,
        char_field='Word',
        date_field=now(),
        date_time_field=now(),
        decimal_field=19.88,
        duration_field=datetime.timedelta(days=1, minutes=60),
        email_field='me@here.com',
        file_path_field='static.css',
        float_field=88.19,
        generic_ip_address_field='127.0.01',
        ip_address_field='0.0.0.0',
        integer_field=-23,
        positive_big_integer_field=3247238947238,
        positive_integer_field=1312321,
        positive_small_integer_field=9,
        slug_field='',
        small_integer_field=-7,
        text_field='This is a bunch of text in the text field that is written to take up a lot of space',
        time_field=datetime.time(10, 20, 30),
        url_field='https://www.stratusadv.com',
        uuid_field=uuid.uuid4()
    )

    exclude_test_model = BigTestModel.objects.all().order_by('-id')[:limit]
    BigTestModel.objects.exclude(pk__in=exclude_test_model).delete()

    return big_test_model



import inspect
from abc import ABC, abstractmethod
from typing import Optional

from django_glue.enums import GlueAction, GlueAccess
from django_glue.data_classes import GlueBodyData, GlueJsonResponseData
from django_glue.responses import generate_json_404_response_data


class Service(ABC):
    def process_body_data(self, access: GlueAccess, body_data: GlueBodyData) -> GlueJsonResponseData:
        action = body_data.action
        if access.has_access(action.required_access):
            if action == GlueAction.GET:
                return self.process_get_action(body_data)
            if action == GlueAction.CREATE:
                return self.process_create_action(body_data)
            if action == GlueAction.UPDATE:
                return self.process_update_action(body_data)
            if action == GlueAction.DELETE:
                return self.process_delete_action(body_data)
            if action == GlueAction.METHOD:
                return self.process_method_action(body_data)
            else:
                return generate_json_404_response_data()
        else:
            return generate_json_404_response_data()

    @abstractmethod
    def process_get_action(self, body_data: GlueBodyData) -> GlueJsonResponseData:
        return generate_json_404_response_data()

    def process_create_action(self, body_data: GlueBodyData) -> GlueJsonResponseData:
        return generate_json_404_response_data()

    def process_update_action(self, body_data: GlueBodyData) -> GlueJsonResponseData:
        return generate_json_404_response_data()

    def process_delete_action(self, body_data: GlueBodyData) -> GlueJsonResponseData:
        return generate_json_404_response_data()

    def process_method_action(self, body_data: GlueBodyData) -> GlueJsonResponseData:
        return generate_json_404_response_data()




from dataclasses import fields

from django.core import serializers


def camel_to_snake(string):
    return ''.join(['_' + c.lower() if c.isupper() else c for c in string]).lstrip('_')


def remove_none_value_field_from_data_class_object(obj):
    for field in fields(obj):
        if getattr(obj, field.name) is None:
            delattr(obj, field.name)


def serialize_object_to_json(model_object):
    return serializers.serialize('json', [model_object, ])



import json

from django.shortcuts import HttpResponse

from django_glue.handlers import GlueDataRequestHandler
from django_glue.sessions import GlueKeepLiveSession


def glue_data_ajax_handler_view(request):
    return GlueDataRequestHandler(request).process_response()


def glue_keep_live_handler_view(request):
    data = json.loads(request.body.decode('utf-8'))
    unique_names = data['unique_names']

    GlueKeepLiveSession(request).update_unique_names(unique_names)

    return HttpResponse(status=204)



from typing import Optional

from django.http import JsonResponse

from django_glue.data_classes import GlueJsonResponseData, GlueJsonData
from django_glue.enums import GlueJsonResponseType, GlueJsonResponseStatus


def generate_json_200_response(
        message_title: str,
        message_body: str,
        data: Optional[GlueJsonData] = None,
        optional_message_data: Optional[dict] = None,
) -> JsonResponse:

    return generate_json_200_response_data(
        message_title=message_title,
        message_body=message_body,
        data=data,
        optional_message_data=optional_message_data,
    ).to_django_json_response()


def generate_json_200_response_data(
        message_title: str,
        message_body: str,
        data: Optional[GlueJsonData] = None,
        optional_message_data: Optional[dict] = None,
) -> GlueJsonResponseData:

    return GlueJsonResponseData(
        message_title=message_title,
        message_body=message_body,
        data=data,
        optional_message_data=optional_message_data,
        response_status=GlueJsonResponseStatus.SUCCESS,
        response_type=GlueJsonResponseType.SUCCESS,
    )


def generate_json_404_response(
        message_title: str = 'Request not Found',
        message_body: str = 'The requested information, object or view you are looking for was not found.',
) -> JsonResponse:
    return generate_json_404_response_data(
        message_title,
        message_body,
    ).to_django_json_response()


def generate_json_404_response_data(
        message_title: str = 'Request not Found',
        message_body: str = 'The requested information, object or view you are looking for was not found.',
) -> GlueJsonResponseData:
    return GlueJsonResponseData(
        message_title=message_title,
        message_body=message_body,
        response_status=GlueJsonResponseStatus.ERROR,
        response_type=GlueJsonResponseType.ERROR,
    )



import logging

from django_glue.data_classes import GlueBodyData, GlueMetaData
from django_glue.services.services import Service
from django.shortcuts import HttpResponse
from django.template.loader import render_to_string


class GlueTemplateService(Service):
    def __init__(self, meta_data: GlueMetaData) -> None:
        self.meta_data = meta_data

    def process_get_action(self, body_data: GlueBodyData) -> HttpResponse:
        return HttpResponse(render_to_string(self.meta_data.template, body_data.data['data']))



# Todo: these setting are not fully implemented through out all of the code

DJANGO_GLUE_CONTEXT_NAME = 'django_glue'
DJANGO_GLUE_SESSION_NAME = 'django_glue'

DJANGO_GLUE_KEEP_LIVE_CONTEXT_NAME = 'django_glue_keep_live'
DJANGO_GLUE_KEEP_LIVE_SESSION_NAME = 'django_glue_keep_live'

DJANGO_GLUE_KEEP_LIVE_EXPIRE_TIME_SECONDS = 120.0
DJANGO_GLUE_KEEP_LIVE_INTERVAL_TIME_SECONDS = 90.0

DJANGO_GLUE_AUTO_DELETE = False

DJANGO_GLUE_URL = 'django_glue/'



from __future__ import annotations

import json
from typing import Any, Optional, Union
from dataclasses import dataclass, asdict

from django.contrib.contenttypes.models import ContentType
from django.http import JsonResponse

from django_glue.enums import GlueConnection, GlueAccess, GlueAction, GlueJsonResponseStatus, GlueJsonResponseType
from django_glue.core.utils import remove_none_value_field_from_data_class_object


class GlueBodyData:
    def __init__(self, request_body):
        self.data = json.loads(request_body.decode('utf-8'))
        self.glue_action = GlueAction(self.data['action'])

    def __getitem__(self, key):
        return self.data[key]

    def __setitem__(self, key, value):
        self.data[key] = value

    @property
    def action(self) -> GlueAction:
        return self.glue_action

    @property
    def unique_name(self) -> str:
        return self.data['unique_name']


@dataclass
class GlueContextData:
    connection: GlueConnection
    access: GlueAccess = None
    fields: dict[GlueModelFieldData] = None
    methods: list[str] = None

    def to_dict(self) -> dict:
        return asdict(self)



@dataclass
class GlueJsonData:
    fields: Optional[dict] = None
    simple_fields: Optional[dict] = None
    method_return: Optional[Any] = None
    custom: Optional[dict] = None

    def to_dict(self):
        remove_none_value_field_from_data_class_object(self)
        return asdict(self)


@dataclass
class GlueJsonResponseData:
    message_title: Optional[str] = None
    message_body: Optional[str] = None
    data: Optional[GlueJsonData] = None
    optional_message_data: Optional[dict] = None
    response_type: GlueJsonResponseType = GlueJsonResponseType.SUCCESS
    response_status: GlueJsonResponseStatus = GlueJsonResponseStatus.SUCCESS

    def to_dict(self) -> dict:
        remove_none_value_field_from_data_class_object(self)

        json_response_dict = asdict(self)

        if isinstance(self.data, GlueJsonData):
            json_response_dict['data'] = self.data.to_dict()

        return json_response_dict

    def to_django_json_response(self) -> JsonResponse:
        return JsonResponse(self.to_dict(), status=self.response_status.value)


@dataclass
class GlueModelFieldData:
    type: str
    value: Any
    html_attr: dict

    def to_dict(self) -> dict:
        return asdict(self)


@dataclass
class GlueMetaData:
    app_label: str = None
    model: str = None
    object_pk: int = None
    query_set_str: str = None
    template: str = None
    fields: Union[list, tuple] = None
    exclude: Union[list, tuple] = None
    methods: Union[list, tuple] = None

    @property
    def model_class(self):
        return ContentType.objects.get_by_natural_key(self.app_label, self.model).model_class()

    def to_dict(self) -> dict:
        return asdict(self)



import inspect
from typing import Optional

from django.contrib.contenttypes.models import ContentType
from django.db.models import Model

from django_glue.data_classes import GlueJsonResponseData, GlueJsonData, GlueBodyData, GlueMetaData
from django_glue.responses import generate_json_200_response_data, generate_json_404_response_data
from django_glue.services.services import Service
from django_glue.utils import generate_simple_field_dict, get_fields_from_model, check_valid_method_kwargs, \
    type_set_method_kwargs, field_name_included


class GlueModelObjectService(Service):
    def __init__(self, meta_data: GlueMetaData) -> None:
        self.meta_data = meta_data
        self.model_class = None
        self.object: Optional[Model] = None

    def load_object(self):
        self.load_model_class()
        self.object = self.model_class.objects.get(pk=self.meta_data.object_pk)

    def load_model_class(self):
        self.model_class = ContentType.objects.get_by_natural_key(
            self.meta_data.app_label,
            self.meta_data.model
        ).model_class()

    def process_get_action(self, body_data: GlueBodyData) -> GlueJsonResponseData:
        self.load_object()
        json_data = GlueJsonData()

        json_data.simple_fields = generate_simple_field_dict(
            self.object,
            self.meta_data.fields,
            self.meta_data.exclude,
        )

        # json_data.simple_fields['first_name'] = 'Humbugery'

        return generate_json_200_response_data(
            'THE GET ACTION',
            'this is a response from an model object get action!!! stay tuned!',
            json_data,
        )

    def process_create_action(self, body_data: GlueBodyData) -> GlueJsonResponseData:
        self.load_model_class()

        model_object = self.model_class()

        for field in get_fields_from_model(model_object):
            if field.name in body_data['data'] and field.name != 'id':
                model_object.__dict__[field.name] = body_data['data'][field.name]

        model_object.save()

        json_data = GlueJsonData()

        json_data.simple_fields = generate_simple_field_dict(
            model_object,
            self.meta_data.fields,
            self.meta_data.exclude
        )

        return generate_json_200_response_data(
            'THE CREATE ACTION',
            'this is a response from an model object create action!',
            json_data
        )

    def process_update_action(self, body_data: GlueBodyData) -> GlueJsonResponseData:
        self.load_object()

        for field in get_fields_from_model(self.object):
            if field.name in body_data['data'] and field.name != 'id' and field_name_included(field.name, self.meta_data.fields, self.meta_data.exclude):
                self.object.__dict__[field.name] = body_data['data'][field.name]

        self.object.save()

        return generate_json_200_response_data(
            'THE UPDATE ACTION',
            'this is a response from an model object update action!'
        )

    def process_delete_action(self, body_data: GlueBodyData) -> GlueJsonResponseData:
        self.load_object()
        self.object.delete()

        return generate_json_200_response_data(
            'THE DELETE ACTION',
            'this is a response from an model object delete action!'
        )

    def process_method_action(self, body_data: GlueBodyData) -> GlueJsonResponseData:
        self.load_object()
        kwargs = body_data['data']['kwargs']
        method_return = None

        if body_data['data']['method'] in self.meta_data.methods and hasattr(self.model_class, body_data['data']['method']):
            method = getattr(self.object, body_data['data']['method'])

            if check_valid_method_kwargs(method, kwargs):
                type_set_kwargs = type_set_method_kwargs(method, kwargs)

                method_return = method(**type_set_kwargs)
            else:
                return generate_json_404_response_data()
        else:
            return generate_json_404_response_data()

        json_data = GlueJsonData()

        json_data.method_return = method_return

        return generate_json_200_response_data(
            'THE METHOD ACTION',
            'this is a response from an model object method action!',
            json_data
        )





from typing import Union

from django.db.models import Model
from django.db.models.query import QuerySet
from django.http import HttpRequest

from django_glue.sessions import GlueSession, GlueKeepLiveSession


def add_glue(
        request: HttpRequest,
        unique_name: str,
        target: Union[Model, QuerySet, str],
        access: str = 'view',
        fields: Union[list, tuple] = ('__all__',),
        exclude: Union[list, tuple] = ('__none__',),
        methods: Union[list, tuple] = ('__none__',),
) -> None:
    if isinstance(fields, (list, tuple)) and isinstance(exclude, (list, tuple)):
        glue_session = GlueSession(request)

        if isinstance(target, Model):
            glue_session.add_model_object(unique_name, target, access, fields, exclude, methods)

        elif isinstance(target, QuerySet):
            glue_session.add_query_set(unique_name, target, access, fields, exclude, methods)

        elif isinstance(target, str):
            glue_session.add_template(unique_name, target)

        else:
            raise f'target is not a valid type must be a django.db.models.Model or django.db.models.query.QuerySet'

        glue_keep_live_session = GlueKeepLiveSession(request)

        glue_keep_live_session.set_unique_name(unique_name)

        glue_session.set_modified()

    else:
        raise f'fields or exclude is not a valid type must be a list or tuple'



from typing import Optional

from django.db.models import QuerySet

from django_glue.data_classes import GlueJsonResponseData, GlueBodyData, GlueMetaData, GlueJsonData
from django_glue.responses import generate_json_200_response_data, generate_json_404_response_data
from django_glue.services.services import Service
from django_glue.utils import decode_query_set_from_str, generate_simple_field_dict, get_fields_from_model, \
    check_valid_method_kwargs, type_set_method_kwargs, field_name_included


class GlueQuerySetService(Service):
    def __init__(self, meta_data: GlueMetaData) -> None:
        self.meta_data = meta_data
        self.query_set: Optional[QuerySet] = None

    def load_query_set(self):
        self.query_set = decode_query_set_from_str(self.meta_data.query_set_str)

    def process_get_action(self, body_data: GlueBodyData) -> GlueJsonResponseData:
        self.load_query_set()

        if 'filter_params' in body_data['data']:
            data = []
            for model_object in self.query_set.filter(**body_data['data']['filter_params']):
                data.append(GlueJsonData(simple_fields=generate_simple_field_dict(
                        model_object, self.meta_data.fields, self.meta_data.exclude)))

        elif 'all' in body_data['data']:
            data = []
            for model_object in self.query_set.all():
                data.append(GlueJsonData(simple_fields=generate_simple_field_dict(
                        model_object, self.meta_data.fields, self.meta_data.exclude)))

        else:
            model_object = self.query_set.get(id=body_data['data']['id'])
            data = GlueJsonData(simple_fields=generate_simple_field_dict(
                model_object, self.meta_data.fields, self.meta_data.exclude))

        return generate_json_200_response_data(
            'THE QUERY GET ACTION',
            'this is a response from an query set get action',
            data
        )

    def process_create_action(self, body_data: GlueBodyData) -> GlueJsonResponseData:
        self.load_query_set()

        model_object = self.meta_data.model_class()

        for field in get_fields_from_model(self.meta_data.model_class):
            if field.name in body_data['data'] and field.name != 'id':
                setattr(model_object, field.name, body_data['data'][field.name])

        model_object.save()

        json_data = GlueJsonData()

        json_data.simple_fields = generate_simple_field_dict(
            model_object,
            self.meta_data.fields,
            self.meta_data.exclude
        )

        return generate_json_200_response_data(
            'THE QUERY CREATE ACTION',
            'this is a response from an query set create action!',
            json_data
        )

    def process_update_action(self, body_data: GlueBodyData) -> GlueJsonResponseData:
        self.load_query_set()

        model_object = self.meta_data.model_class.objects.get(id=body_data['data']['id'])

        for field in get_fields_from_model(self.meta_data.model_class):
            if field.name in body_data['data'] and field.name != 'id' and field_name_included(field.name, self.meta_data.fields, self.meta_data.exclude):
                model_object.__dict__[field.name] = body_data['data'][field.name]

        model_object.save()

        return generate_json_200_response_data(
            'THE QUERY UPDATE ACTION',
            'this is a response from an query set update action!'
        )

    def process_delete_action(self, body_data: GlueBodyData) -> GlueJsonResponseData:
        self.load_query_set()

        if isinstance(body_data['data']['id'], list) or isinstance(body_data['data']['id'], tuple):
            self.query_set.filter(id__in=body_data['data']['id']).delete()

        elif isinstance(body_data['data']['id'], int):
            self.query_set.filter(id=body_data['data']['id']).delete()

        else:
            return generate_json_404_response_data()

        return generate_json_200_response_data(
            'THE QUERY DELETE ACTION',
            'this is a response from an query set delete action!'
        )

    def process_method_action(self, body_data: GlueBodyData) -> GlueJsonResponseData:
        self.load_query_set()

        if isinstance(body_data['data']['id'], list) or isinstance(body_data['data']['id'], tuple):
            model_object = self.query_set.filter(id__in=body_data['data']['id'])

        elif isinstance(body_data['data']['id'], int):
            model_object = self.query_set.get(id=body_data['data']['id'])

        else:
            return generate_json_404_response_data()

        kwargs = body_data['data']['kwargs']
        method_return = None

        if body_data['data']['method'] in self.meta_data.methods and hasattr(self.meta_data.model_class, body_data['data']['method']):
            if isinstance(model_object, QuerySet):
                method_return = []
                for object in model_object:
                    method = getattr(object, body_data['data']['method'])
                    if check_valid_method_kwargs(method, kwargs):
                        type_set_kwargs = type_set_method_kwargs(method, kwargs)
                        method_return.append(method(**type_set_kwargs))

            elif isinstance(model_object, self.meta_data.model_class):
                method = getattr(model_object, body_data['data']['method'])

                if check_valid_method_kwargs(method, kwargs):
                    type_set_kwargs = type_set_method_kwargs(method, kwargs)
                    method_return = method(**type_set_kwargs)

            else:
                return generate_json_404_response_data()
        else:
            return generate_json_404_response_data()

        json_data = GlueJsonData()

        json_data.method_return = method_return

        return generate_json_200_response_data(
            'THE METHOD ACTION',
            'this is a response from an query set method action!',
            json_data
        )


from __future__ import annotations

from enum import Enum


class GlueAccess(str, Enum):
    # The order of these variables controls how the permission cascade each other in the has_access method
    VIEW = 'view'
    ADD = 'add'
    CHANGE = 'change'
    DELETE = 'delete'
    ADMIN = 'admin'

    def __str__(self):
        return self.value

    def has_access(self, access_required: GlueAccess):
        glue_access_tuple = tuple(GlueAccess.__members__.values())
        if glue_access_tuple.index(self) >= glue_access_tuple.index(access_required):
            return True
        else:
            return False


class GlueAction(str, Enum):
    GET = 'get'
    CREATE = 'create'
    UPDATE = 'update'
    DELETE = 'delete'
    METHOD = 'method'

    def __str__(self):
        return self.value

    @property
    def required_access(self) -> GlueAccess:
        if self.value == 'get':
            return GlueAccess.VIEW
        elif self.value == 'create':
            return GlueAccess.ADD
        elif self.value == 'update':
            return GlueAccess.CHANGE
        elif self.value == 'delete':
            return GlueAccess.DELETE
        elif self.value == 'method':
            return GlueAccess.VIEW


class GlueConnection(str, Enum):
    MODEL_OBJECT = 'model_object'
    QUERY_SET = 'query_set'
    TEMPLATE = 'template'

    def __str__(self):
        return self.value


class GlueJsonResponseType(str, Enum):
    SUCCESS = 'success'
    INFO = 'info'
    WARNING = 'warning'
    ERROR = 'error'
    DEBUG = 'debug'

    def __str__(self):
        return self.value


class GlueJsonResponseStatus(str, Enum):
    SUCCESS = '200'
    SILENT_SUCCESS = '204'
    ERROR = '404'

    def __str__(self):
        return self.value




import inspect
import logging, pickle, base64, json
from typing import Optional, Callable

from django.core import exceptions

from django_glue.core.utils import serialize_object_to_json
from django_glue.data_classes import GlueModelFieldData


def decode_query_set_from_str(query_set_string):
    query = pickle.loads(base64.b64decode(query_set_string))
    decoded_query_set = query.model.objects.all()
    decoded_query_set.query = query
    return decoded_query_set


def encode_query_set_to_str(query_set):
    return base64.b64encode(pickle.dumps(query_set.query)).decode()


def field_name_included(name, fields, exclude):
    included = False
    if name not in exclude or exclude[0] == '__none__':
        if name in fields or fields[0] == '__all__':
            included = True

    return included


def generate_field_attr_dict(field):
    form_field = field.formfield()
    return form_field.widget_attrs(form_field.widget)


def generate_field_dict(model_object, fields, exclude):
    fields_dict = dict()

    model = type(model_object)
    json_model = json.loads(serialize_object_to_json(model_object))[0]

    for field in model._meta.fields:
        try:
            if field_name_included(field.name, fields, exclude):
                if hasattr(field, 'get_internal_type'):
                    if field.name == 'id':
                        field_value = json_model['pk']
                        field_attr = ''
                    else:
                        field_value = json_model['fields'][field.name]
                        field_attr = generate_field_attr_dict(field)

                    fields_dict[field.name] = GlueModelFieldData(
                        type=field.get_internal_type(),
                        value=field_value,
                        html_attr=field_attr,
                    ).to_dict()

        except:
            raise f'Field "{field.name}" is invalid field or exclude for model type "{model.__class__.__name__}"'

    return fields_dict


def generate_method_list(model_object, methods: tuple) -> list:
    methods_list = list()

    model = type(model_object)

    if methods[0] != '__none__':
        for method in methods:
            if hasattr(model_object, method):
                methods_list.append(method)
            else:
                raise f'Method "{method}" is invalid for model type "{model.__class__.__name__}"'

    return methods_list


def generate_simple_field_dict(model_object, fields, exclude):
    fields_dict = generate_field_dict(model_object, fields, exclude)
    simple_fields_dict = {}

    for key, val in fields_dict.items():
        simple_fields_dict[key] = val['value']

    return simple_fields_dict


def get_fields_from_model(model):
    return [field for field in model._meta.fields]


def check_valid_method_kwargs(method: Callable, kwargs: Optional[dict]):
    for kwarg in kwargs:
        if kwarg not in inspect.signature(method).parameters.keys():
            return False
    return True


def type_set_method_kwargs(method: Callable, kwargs: Optional[dict]) -> dict:
    type_set_kwargs = {}

    # This is a dict consisting of all kwargs and there type annotations (If they have type annotations)
    annotations = inspect.getfullargspec(method).annotations

    for kwarg in kwargs:
        if kwarg in annotations:
            # Converts the kwarg to match the type specified in on the methods kwargs
            type_set_kwargs[kwarg] = inspect.getfullargspec(method).annotations[kwarg](kwargs[kwarg])
        else:
            # If there is not a type annotation, the value remains the same
            type_set_kwargs[kwarg] = kwargs[kwarg]

    return type_set_kwargs


from django.contrib import admin
from django.urls import path, include
from django.views.generic import TemplateView

from tests import views

urlpatterns = [
    path("", views.ModelObjectView.as_view(), name="model_object"),
    path("query_set/", views.QuerySetView.as_view(), name="query_set"),
    path("other/", views.OtherView.as_view(), name="other_glue"),
    path("no_glue/", TemplateView.as_view(template_name='page/no_glue_page.html'), name="no_glue"),
    path("benchmark/", views.benchmark_view, name="benchmark"),
    path("django_glue/", include('django_glue.urls', namespace='django_glue')),
    path("template/", views.template_view, name="template"),
    path("view/", views.view_view, name="view"),
    path("view/card/", views.view_card_view, name="view_card"),
]

urlpatterns += [
    path('admin/', admin.site.urls),
]


import json

from django import template
from django.utils.safestring import mark_safe

from django_glue.conf import settings

register = template.Library()


@register.inclusion_tag('django_glue/django_glue.html', takes_context=True)
def glue_init(context):
    return context


@register.simple_tag(takes_context=True)
def glue_context_data_str(context):
    return mark_safe(str(json.dumps(context[settings.DJANGO_GLUE_CONTEXT_NAME])))


@register.simple_tag(takes_context=True)
def glue_context_keep_live_str(context):
    return mark_safe(str(json.dumps(context[settings.DJANGO_GLUE_KEEP_LIVE_CONTEXT_NAME])))


@register.tag
def glue_html_attr(parser, token):
    try:
        tokens = token.contents.split(None, 1)
        unique_name, field = tokens[1].split('.', 1)
    except ValueError:
        raise template.TemplateSyntaxError(f'{token.contents.split()[0]} tag requires arguments')

    return GlueHTMLAttrNode(unique_name, field)


class GlueHTMLAttrNode(template.Node):
    def __init__(self, unique_name, field):
        self.unique_name = unique_name
        self.field = field

    def render(self, context):
        html_attr_str = ''

        for key, val in context[settings.DJANGO_GLUE_CONTEXT_NAME]['context'][self.unique_name]['fields'][self.field]['html_attr'].items():
            html_attr_str = f'{key}="{val}" '

        return html_attr_str


from django.conf import settings as user_settings
from django_glue import settings as default_settings


class DjangoGlueSettings:
    def __getattr__(self, name):
        if hasattr(user_settings, name):
            return getattr(user_settings, name)

        if hasattr(default_settings, name):
            return getattr(default_settings, name)

        raise f'No attribute {name} found in settings.'


settings = DjangoGlueSettings()


from setuptools import find_packages, setup

from django_glue import __version__

setup(
    name="django-glue",
    version=__version__,
    description="Industrial Strength Glue for Django Backends and Frontends!",
    long_description=open("README.md").read(),
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Environment :: Web Environment",
        "Framework :: Django",
        "Framework :: Django :: 3.2",
        "Framework :: Django :: 4.0",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Programming Language :: JavaScript",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Topic :: Internet :: WWW/HTTP",
        "Topic :: Internet :: WWW/HTTP :: Dynamic Content",
        "Topic :: Software Development :: Libraries :: Python Modules",
    ],
    keywords=["glue", "django", "backend", "frontend", "javascript", "active server pages"],
    author="Nathan Johnson, Austin Sauer & Wesley Howery",
    author_email="info@stratusadv.com",
    url="https://github.com/stratusadv/django-glue",
    license="MIT",
    packages=find_packages(exclude=["docs"]),
    include_package_data=True,
    zip_safe=False,
    python_requires=">=3.9",
    install_requires="django>=3.2",
)




import logging

from django.template.response import TemplateResponse
from django.views.generic import TemplateView

from tests.models import TestModel
from tests.utils import generate_randomized_test_model, generate_big_test_model
from django_glue.glue import add_glue


class ModelObjectView(TemplateView):
    template_name = 'page/model_object_page.html'

    def get_context_data(self, **kwargs):
        context_data = super().get_context_data(**kwargs)

        test_model_object = generate_randomized_test_model()

        add_glue(self.request, 'test_model_1', test_model_object, 'delete', exclude=('birth_date', 'anniversary_datetime'), methods=['is_lighter_than', 'get_full_name'])
        add_glue(self.request, 'test_model_2', test_model_object, 'change', exclude=('birth_date', 'anniversary_datetime'))
        add_glue(self.request, 'test_model_3', test_model_object, 'delete', exclude=('birth_date', 'anniversary_datetime'))
        add_glue(self.request, 'test_model_4', test_model_object, 'change', exclude=('birth_date', 'anniversary_datetime'))

        big_test_model_object = generate_big_test_model()

        return context_data


class QuerySetView(TemplateView):
    template_name = 'page/query_set_page.html'

    def get_context_data(self, **kwargs):
        context_data = super().get_context_data(**kwargs)

        add_glue(self.request, 'test_query_1', TestModel.objects.filter(id__gte=1).filter(id__lte=10000), 'delete', exclude=('birth_date', 'anniversary_datetime'), methods=['is_lighter_than', 'get_full_name'])
        add_glue(self.request, 'test_query_2', TestModel.objects.filter(id__gte=1).filter(id__lte=10000), 'add', exclude=('birth_date', 'anniversary_datetime'))
        add_glue(self.request, 'test_query_3', TestModel.objects.filter(id__gte=1).filter(id__lte=10000), 'change', exclude=('birth_date', 'anniversary_datetime'))

        return context_data


class OtherView(TemplateView):
    template_name = 'page/other_glue_page.html'
    def get_context_data(self, **kwargs):
        context_data = super().get_context_data(**kwargs)

        other_test_model_object = generate_randomized_test_model()

        logging.warning(f'Added Other TestModel object.')

        add_glue(self.request, 'other_test_model_1', other_test_model_object, 'change', exclude=('birth_date', 'anniversary_datetime',))

        logging.warning('Added model object glue for Other TestModel Object in write mode')

        return context_data


def benchmark_view(request):
    from time import time
    import random
    import string

    benchmarks = {}

    test_model_object = generate_randomized_test_model()

    def generate_random_string(length):
        return ''.join(random.choice(string.ascii_letters) for _ in range(length))

    start = time()
    stop = time()

    benchmarks['No Glue'] = stop - start

    random_strings = [generate_random_string(5) for _ in range(1)]

    start = time()
    for unique_name in random_strings:
        add_glue(request, unique_name, test_model_object, 'delete', exclude=('birth_date', 'anniversary_datetime'), methods=['is_lighter_than', 'get_full_name'])
    stop = time()

    benchmarks['1 Glue'] = stop - start

    random_strings = [generate_random_string(5) for _ in range(10)]

    start = time()
    for unique_name in random_strings:
        add_glue(request, unique_name, test_model_object, 'delete', exclude=('birth_date', 'anniversary_datetime'), methods=['is_lighter_than', 'get_full_name'])
    stop = time()

    benchmarks['10 Glue'] = stop - start

    random_strings = [generate_random_string(5) for _ in range(100)]

    start = time()
    for unique_name in random_strings:
        add_glue(request, unique_name, test_model_object, 'delete', exclude=('birth_date', 'anniversary_datetime'), methods=['is_lighter_than', 'get_full_name'])
    stop = time()

    benchmarks['100 Glue'] = stop - start

    random_strings = [generate_random_string(5) for _ in range(1000)]

    start = time()
    for unique_name in random_strings:
        add_glue(request, unique_name, test_model_object, 'delete', exclude=('birth_date', 'anniversary_datetime'), methods=['is_lighter_than', 'get_full_name'])
    stop = time()

    benchmarks['1000 Glue'] = stop - start

    return TemplateResponse(request, 'page/benchmark_page.html', {'benchmarks': benchmarks})


def view_view(request):
    return TemplateResponse(request, 'page/view_page.html')


def view_card_view(request):
    test_model_object = generate_randomized_test_model()

    add_glue(request, 'test_model_view_card', test_model_object, 'delete', exclude=('birth_date', 'anniversary_datetime'),
             methods=['is_lighter_than', 'get_full_name'])

    return TemplateResponse(request, 'card/view_card.html')

def template_view(request):
    add_glue(request, 'button_1', 'element/button_element.html')

    return TemplateResponse(request, 'page/template_page.html')


from time import time

from django.contrib.contenttypes.models import ContentType

from django_glue.conf import settings
from django_glue.enums import GlueConnection, GlueAccess
from django_glue.data_classes import GlueContextData, GlueMetaData
from django_glue.utils import generate_field_dict, generate_method_list, encode_query_set_to_str

GLUE_SESSION_TYPES = (
    'context',
    'meta',
)


class GlueSession:
    def __init__(self, request):
        self.request = request

        self.request.session.setdefault(settings.DJANGO_GLUE_SESSION_NAME, dict())

        for session_type in GLUE_SESSION_TYPES:
            self.request.session[settings.DJANGO_GLUE_SESSION_NAME].setdefault(session_type, dict())

        self.session = self.request.session[settings.DJANGO_GLUE_SESSION_NAME]

    def __getitem__(self, key):
        return self.session[key]

    def __setitem__(self, key, value):
        self.session[key] = value

    def add_model_object(
            self,
            unique_name: str,
            model_object,
            access: str = 'view',
            fields: tuple = ('__all__',),
            exclude: tuple = ('__none__',),
            methods: tuple = ('__none__',),
    ):
        content_type = ContentType.objects.get_for_model(model_object)

        self.check_unique_name(unique_name)

        self.add_context(unique_name, GlueContextData(
            connection=GlueConnection('model_object'),
            access=GlueAccess(access),
            fields=generate_field_dict(model_object, fields, exclude),
            methods=generate_method_list(model_object, methods),
        ))

        self.add_meta(unique_name, GlueMetaData(
            app_label=content_type.app_label,
            model=content_type.model,
            object_pk=model_object.pk,
            fields=fields,
            exclude=exclude,
            methods=methods,
        ))

        self.set_modified()

    def add_query_set(
            self,
            unique_name: str,
            query_set,
            access: str = 'view',
            fields: tuple = ('__all__',),
            exclude: tuple = ('__none__',),
            methods: tuple = ('__none__',),
    ):
        content_type = ContentType.objects.get_for_model(query_set.query.model)

        self.check_unique_name(unique_name)

        self.add_context(unique_name, GlueContextData(
            connection=GlueConnection('query_set'),
            access=GlueAccess(access),
            fields=generate_field_dict(query_set.query.model(), fields, exclude),
            methods=generate_method_list(query_set.query.model(), methods),
        ))

        self.add_meta(unique_name, GlueMetaData(
            app_label=content_type.app_label,
            model=content_type.model,
            query_set_str=encode_query_set_to_str(query_set),
            fields=fields,
            exclude=exclude,
            methods=methods,
        ))

        self.set_modified()

    def add_context(self, unique_name, context_data: GlueContextData) -> None:
        self.session['context'][unique_name] = context_data.to_dict()

    def add_template(
            self,
            unique_name: str,
            template,
    ):

        self.check_unique_name(unique_name)

        self.add_context(unique_name, GlueContextData(
            connection=GlueConnection('template'),
            access=GlueAccess('view'),
        ))

        self.add_meta(unique_name, GlueMetaData(
            template=template,
        ))

        self.set_modified()

    def add_meta(self, unique_name, meta_data: GlueMetaData) -> None:
        self.session['meta'][unique_name] = meta_data.to_dict()

    def check_unique_name(self, unique_name):
        if self.unique_name_unused(unique_name):
            self.purge_unique_name(unique_name)

    def clean(self, removable_unique_names):
        for unique_name in removable_unique_names:
            self.purge_unique_name(unique_name)

        self.set_modified()

    def purge_unique_name(self, unique_name):
        for session_type in GLUE_SESSION_TYPES:
            if unique_name in self.session[session_type]:
                self.session[session_type].pop(unique_name)

    def set_modified(self):
        self.request.session.modified = True

    def unique_name_unused(self, unique_name):
        if unique_name in self.session['context']:
            return False
        else:
            return True


class GlueKeepLiveSession:
    def __init__(self, request):
        self.request = request
        self.session = request.session.setdefault(settings.DJANGO_GLUE_KEEP_LIVE_SESSION_NAME, dict())

    def __getitem__(self, key):
        return self.session[key]

    def __setitem__(self, key, value):
        self.session[key] = value

    def clean_and_get_expired_unique_names(self) -> list:
        expired_unique_names = []

        for key, val in self.session.items():
            if time() > val:
                expired_unique_names.append(key)

        for expired_unique_name in expired_unique_names:
            self.session.pop(expired_unique_name)

        self.set_modified()

        return expired_unique_names

    @staticmethod
    def get_next_expire_time():
        return time() + settings.DJANGO_GLUE_KEEP_LIVE_EXPIRE_TIME_SECONDS

    def set_unique_name(self, unique_name):
        self.session.setdefault(unique_name, self.get_next_expire_time())

        self.set_modified()

    def set_modified(self):
        self.request.session.modified = True

    def update_unique_names(self, unique_names: list):
        for unique_name in unique_names:
            if unique_name in self.session:
                self.session[unique_name] = self.get_next_expire_time()

        self.set_modified()



import logging, json

from django_glue.services import GlueModelObjectService, GlueQuerySetService
from django_glue.responses import generate_json_404_response
from django_glue.services.templates import GlueTemplateService
from django_glue.sessions import GlueSession
from django_glue.data_classes import GlueMetaData, GlueBodyData
from django_glue.enums import GlueConnection, GlueAccess


class GlueDataRequestHandler:
    def __init__(self, request):
        self.request = request

        self.is_valid_request = True

        if request.content_type == 'application/json' or request.content_type == 'text/html':
            logging.warning(request.body.decode('utf-8'))
            self.body_data = GlueBodyData(request.body)
        else:
            self.is_valid_request = False

        self.unique_name = self.body_data['unique_name']

        self.glue_session = GlueSession(request)
        self.context = self.glue_session['context']

        if self.unique_name in self.context:
            self.method = request.method

            self.meta_data: GlueMetaData = GlueMetaData(
                **self.glue_session['meta'][self.unique_name]
            )

            self.connection = GlueConnection(self.context[self.unique_name]['connection'])

            self.access = GlueAccess(self.context[self.unique_name]['access'])

        else:
            self.is_valid_request = False

        logging.warning(f'{self.is_valid_request = }')

    def process_response(self):
        if self.method == 'QUERY' and self.is_valid_request:
            if self.connection == GlueConnection.MODEL_OBJECT:
                glue_model_object_service = GlueModelObjectService(
                    self.meta_data,
                )

                json_response_data = glue_model_object_service.process_body_data(self.access, self.body_data)
                return json_response_data.to_django_json_response()

            elif self.connection == GlueConnection.QUERY_SET:
                glue_query_set_service = GlueQuerySetService(
                    self.meta_data,
                )

                json_response_data = glue_query_set_service.process_body_data(self.access, self.body_data)
                return json_response_data.to_django_json_response()

            elif self.connection == GlueConnection.TEMPLATE:
                glue_template_service = GlueTemplateService(
                    self.meta_data,
                )

                html_response_data = glue_template_service.process_body_data(self.access, self.body_data)
                return html_response_data

            else:
                return generate_json_404_response()

        else:
            return generate_json_404_response()



from django.db import models
from django.utils.timezone import now

from django_glue.data_classes import GlueJsonResponseData


class TestModel(models.Model):
    first_name = models.CharField(max_length=32)
    last_name = models.CharField(max_length=32)
    description = models.TextField()
    favorite_number = models.IntegerField()
    anniversary_datetime = models.DateTimeField(default=now)
    birth_date = models.DateField(default=now)
    weight_lbs = models.DecimalField(max_digits=7, decimal_places=3)

    def __str__(self):
        return f'{self.first_name} {self.last_name}'

    def is_lighter_than(self, check_weight: float) -> bool:
        if self.weight_lbs < check_weight:
            return True
        return False

    def get_full_name(self):
        return f'{self.first_name} {self.last_name}'


class BigTestModel(models.Model):
    big_integer_field = models.BigIntegerField()
    binary_field = models.BinaryField()
    boolean_field = models.BooleanField()
    char_field = models.CharField(max_length=8)
    date_field = models.DateField()
    date_time_field = models.DateTimeField()
    decimal_field = models.DecimalField(decimal_places=2, max_digits=6)
    duration_field = models.DurationField()
    email_field = models.EmailField()
    file_path_field = models.FilePathField(path='/static')
    float_field = models.FloatField()
    generic_ip_address_field = models.GenericIPAddressField()
    ip_address_field = models.GenericIPAddressField()
    integer_field = models.IntegerField()
    positive_big_integer_field = models.PositiveBigIntegerField()
    positive_integer_field = models.PositiveIntegerField()
    positive_small_integer_field = models.PositiveSmallIntegerField()
    slug_field = models.SlugField()
    small_integer_field = models.SmallIntegerField()
    text_field = models.TextField()
    time_field = models.TimeField()
    url_field = models.URLField()
    uuid_field = models.UUIDField()

    def __str__(self):
        return f'{self.char_field}'



